/*
 * sverb: an order 15 CFDN reverb
 * http://sed.free.fr/sverb
 * public domain
 * (there may be some patent for this stuff, I don't know and I don't care)
 * Cdric Roux <sed@free.fr>
 *
 * Algorithm comes from:
 * "Maximally Diffusive Yet Efficient Feedback Delay Networks for
 *  Artificial Reverberation", Davide Rocchesso, IEEE SIGNAL
 * PROCESSING LETTERS, VOL. 4, NO. 9, SEPTEMBER 1997
 * http://www.scienze.univr.it/documenti//ArticoloRivista/allegato/
 *   allegato128605.pdf
 * And:
 * "Circulant and elliptic feedback delay networks for artificial
 *  reverberation", D. Rocchesso and J. O. Smith, IEEE Trans. Speech
 * Audio Processing, vol. 5, pp. 51 63, Jan. 1997.
 * http://ccrma.stanford.edu/~jos/cfdn/cfdn.pdf
 * And:
 * http://ccrma.stanford.edu/~jos/Reverb/First_Order_Delay_Filter_Design.html
 * http://ccrma.stanford.edu/~jos/Reverb/Tonal_Correction_Filter.html
 * for the lowpass filters
 */

/* This program tries to guess good values for the delays of the diffusion
 * matrix. It is far from good. If you know a good way to find these
 * parameters automatically, don't hesitate to contribute!
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "sverb.h"

/* return 1 if prime, 0 if not (we should do something faster, no?) */
static int is_prime(int x)
{
  int i;
  int z = sqrt(x)+1;
  for (i=2; i<=z; i++)
    if ((x % i) == 0) return 0;
  return 1;
}

static int primify(int d)
{
  if (d <= 3) return d;
  if (!(d&1)) d++;
  while (!is_prime(d)) d+=2;
  return d;
}

void do_first_block(sverb *s, float *b)
{
  int i;
  *b = s->sverb_process(s, 1);
  b++;
  for (i=1; i<512; i++, b++)
    *b = s->sverb_process(s, 0);
}

void do_block(sverb *s, float *b)
{
  int i;
  for (i=0; i<512; i++, b++)
    *b = s->sverb_process(s, 0);
}

float calc_power(float *b)
{
  float ret = 0;
  int i;
  for (i=0; i<512; i++, b++) ret += (*b) * (*b);
  return ret;
}

void calc_quality(sverb *s, int *ret_dec, float *ret_power)
{
  int i;
  float b[512];
  float power=0;
  float old_power=0;
  float cur_power;
  int dec=0;

  sverb_reset(s);

  /* do the first block */
  do_first_block(s, b);
  cur_power = calc_power(b);
  power += cur_power;
  old_power = cur_power;

  /* next 99 blocks will do up to 1s of reverb, we want decaying power
   * for them */
  for (i=0; i<99; i++) {
    do_block(s, b);
    cur_power = calc_power(b);
    if (cur_power < old_power) dec++; else dec--;
    power += cur_power;
    old_power = cur_power;
  }

#if 1
  /* next 490 blocks go up to 5s for the reverb */
  //for (i=0; i<490; i++) {
  for (i=0; i<100; i++) {
    do_block(s, b);
    power += calc_power(b);
  }
#endif

  *ret_dec = dec;
  *ret_power = power;
}

/* return an int between -600 and 1400 */
int randm(void)
{
  return -600+(int) (2001.0*rand()/(RAND_MAX+1.0));
}

/* calculate new params, take the best, return -1 if the best
 * is no change (we reached the best), 0 if some change
 * but if check_changed = 0, always return 0 and modify the 0th element
 * if no best was found
 */
/* OK. Above comment is out of date, read the code to get the real point. */
int _new_params(sverb *s, int check_changed)
{
  int dec[31];
  int inc[31];
  float power[31];
  int best_dec;
  float best_power;
  int changed;
  int best_i;
  int i;
  int turn;

  /* calc new qualities */
  calc_quality(s, &dec[30], &power[30]);

  for (i=0; i<15; i++) {
    int old = s->DELAY[i];

    /* let force all value to be different */
    turn = 0;
    while (1) {
      int rnd;
      int j, more = 0;
      do rnd = randm(); while (s->DELAY[i] + rnd < 1);
      inc[i] = primify(s->DELAY[i] + rnd);
      for (j=0; j<15; j++)
        if (inc[i] == s->DELAY[j]) more = 1;
      if (!more) break;
      turn++;
      /* just in case, we don't want infinite loop */
      if (turn == 100000) break;
    }

    sverb_set_delay(s, i, inc[i]);
    printf("."); fflush(stdout);
    calc_quality(s, &dec[i], &power[i]);
    sverb_set_delay(s, i, old);
  }

  /* pick up the best */
  best_dec = dec[30];
  best_power = power[30];

#if 0
  /* first comparison - the calc stuff is too weird */
  changed = 0;
  best_i = 0; /* for no gcc warning */
  for (i=0; i<15; i++) {
#define calc(power, dec) (power / 2000 + dec * 20)
    if (calc(power[i], dec[i]) >= calc(best_power, best_dec)) {
      best_dec = dec[i];
      best_power = power[i];
      best_i = i;
      changed = 1;
    }
  }

  if (check_changed && !changed) return -1;
  if (!changed) best_i = (int)(15.0*rand()/(RAND_MAX+1.0));
  if (best_i > 14) best_i = 14;

  sverb_set_delay(s, best_i, inc[best_i]);
  return 0;
#endif

  /* second comparison, let dec have the most impact, then power */
  /* dec cannot decrease but power can, maybe never converging */
  /* well, in fact dec can decrease */
  changed = 0;
  best_i = 0; /* for no gcc warning */
  /* first loop, try to increase the dec */
  for (i=0; i<15; i++) {
    if (dec[i] == best_dec) {
      /* same dec => better power? */
      if (power[i] < best_power) continue;
      best_dec = dec[i];
      best_power = power[i];
      best_i = i;
      changed = 1;
    } else if (dec[i] > best_dec) {
      best_dec = dec[i];
      best_power = power[i];
      best_i = i;
      changed = 1;
    }
  }
  if (changed) {
    sverb_set_delay(s, best_i, inc[best_i]);
    return 0;
  }

  /* it did not work, increase the power (so maybe decrease dec) */
  for (i=0; i<15; i++) {
    if (power[i] < best_power) continue;
    best_dec = dec[i];
    best_power = power[i];
    best_i = i;
    changed = 1;
  }

  /* still did not work? give up, bad values */
  if (!changed) return -1;

  sverb_set_delay(s, best_i, inc[best_i]);
  return 0;
}

int new_params(sverb *s)
{
  /* method 0 */
#if 0
  if (_new_params(s, 1) == 0) return 0;
  if (_new_params(s, 1) == 0) return 0;
  return _new_params(s, 0);
#endif
  /* method 1 */
  while (_new_params(s, 0) == -1);
  return 0;
}

int length(sverb *s)
{
  int i;
  int ret=0;
  for (i=0; i<15; i++) ret += s->DELAY[i];
  return ret;
}

int main(int n, char **v)
{
  sverb s;
  int i;
  int total;
  int wanted_length;
  int l;

  if (n != 2) {
    printf("<prog> <expected delay>\n");
    return 1;
  }
  sscanf(v[1], "%d", &wanted_length);

  sverb_init(&s);
  for (i=0; i<15; i++)
    //if (sverb_set_delay(&s, i, 15-i) == -1)
    if (sverb_set_delay(&s, i, i+1) == -1)
      return -1;
  sverb_set_type(&s, SVERB_ASM);
  sverb_set_integer_precision(&s, 14);
  sverb_use_t60(&s, 0);
  sverb_reset(&s);

  while ((l=length(&s)) < wanted_length) {
    printf("\n%d (", l);
    for (i=0; i<14; i++) printf("%d ", s.DELAY[i]); printf("%d", s.DELAY[i]);
    printf(")\n");
    if (new_params(&s) == -1) {
      printf("stability reached, but might not be what you want\n");
      break;
    }
  }

  total = 0;
  printf("\n");
  for (i=0; i<15; i++) {
    total += s.DELAY[i];
    printf("DELAY[%d] = %d\n", i, s.DELAY[i]);
  }
  printf("total delay = %d, you expected %d\n", total, wanted_length);

  return 0;
}
