/*
 * sverb: an order 15 CFDN reverb
 * http://sed.free.fr/sverb
 * public domain
 * (there may be some patent for this stuff, I don't know and I don't care)
 * Cdric Roux <sed@free.fr>
 *
 * Algorithm comes from:
 * "Maximally Diffusive Yet Efficient Feedback Delay Networks for
 *  Artificial Reverberation", Davide Rocchesso, IEEE SIGNAL
 * PROCESSING LETTERS, VOL. 4, NO. 9, SEPTEMBER 1997
 * http://www.scienze.univr.it/documenti//ArticoloRivista/allegato/
 *   allegato128605.pdf
 * And:
 * "Circulant and elliptic feedback delay networks for artificial
 *  reverberation", D. Rocchesso and J. O. Smith, IEEE Trans. Speech
 * Audio Processing, vol. 5, pp. 51 63, Jan. 1997.
 * http://ccrma.stanford.edu/~jos/cfdn/cfdn.pdf
 * And:
 * http://ccrma.stanford.edu/~jos/Reverb/First_Order_Delay_Filter_Design.html
 * http://ccrma.stanford.edu/~jos/Reverb/Tonal_Correction_Filter.html
 * for the lowpass filters
 */

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#include "sverb_version.h"
#include "sverb.h"

#include "soundcard.h"

typedef struct {
  /* gui elements */
  GtkWidget *main_win;
  GtkWidget *combo_preset;
  GtkWidget *total_delay_label;
  GtkWidget *delay_label[15];
  GtkWidget *delay_entry[15];
  GtkWidget *prime_toggle;
  GtkWidget *t60_0_label;
  GtkWidget *t60_0_entry;
  GtkWidget *t60_pi_label;
  GtkWidget *t60_pi_entry;
  GtkWidget *bypass_t60_toggle;
  GtkWidget *float_button;
  GtkWidget *int_button;
#ifdef ASM_IA32
  GtkWidget *asm_button;
#endif
  GtkWidget *int_precision_scale;
  GtkWidget *impulse_drawing_area;
  GtkWidget *impulse_min;
  GtkWidget *impulse_max;
  GtkWidget *update_impulse_button;
  GtkWidget *impulse_length_scale;
  GtkWidget *impulse_play_button;
  GtkWidget *oss_output_entry;
  GtkWidget *input_scale_label;
  GtkWidget *input_scale_entry;
  GtkWidget *output_scale_label;
  GtkWidget *output_scale_entry;
  GtkWidget *help_window;
  int help_window_shown;
  /* set to 1 when something is changed, reset to 0 by load, save, etc. */
  int modified;
  /* link to sverb struct */
  sverb *s;
  /* impulse response display */
  int min[220];
  int max[220];
  float vmin;
  float vmax;
  int aborted;
  int update_running;
  /* audio */
  int playing;
  unsigned int audio_callback;
  float iscale;
  float oscale;
  /* edited file */
  char file[FILENAME_MAX+1];
} sverb_gui;

/* audio stuff */

/* forward declarations of these functions */
static void impulse_play_callback(GtkButton *, sverb_gui *);
static void alert(void *, char *);

static short buf[4096];
static void audio_ready(sverb_gui *sg, int unused, GdkInputCondition unused2)
{
#define store(y,p) \
  do { \
    if (y<-32768) y = -32768; \
    if (y>32767) y = 32767; \
    buf[p] = y; \
  } while(0)

  float y;
  int i;

  /* is it first call? */
  if (sg->playing & 2) {
    sg->playing = 1;
    y = sg->s->sverb_process(sg->s, sg->iscale);
    y *= sg->oscale;
    store(y, 0);
    for (i=1; i<4096; i++) {
      y = sg->s->sverb_process(sg->s, 0);
      y *= sg->oscale;
      store(y, i);
    }
  } else {
    /* not first call */
    for (i=0; i<4096; i++) {
      y = sg->s->sverb_process(sg->s, 0);
      y *= sg->oscale;
      store(y, i);
    }
  }
  if (!play_sound(buf, 4096*2)) {
    impulse_play_callback(GTK_BUTTON(sg->impulse_play_button), sg);
    alert(sg->main_win, "Audio error, sorry.");
  }
}

/* various functions */

static void calculate_response(sverb_gui *sg)
{
#define inc(k) \
  do { \
    k++; \
    if (k&4095) break; \
    gtk_widget_queue_draw(sg->impulse_drawing_area); \
    while (gtk_events_pending()) \
      gtk_main_iteration(); \
    if (sg->aborted) \
      goto done; \
  } while (0)

  float y;
  int i;
  int j;
  int n;
  int k = 0;
  float vmin;
  float vmax;
  float min[220];
  float max[220];
  float totmin;
  float totmax;

  sg->aborted = 0;
  sg->update_running = 1;

  gtk_button_set_label(GTK_BUTTON(sg->update_impulse_button), "Stop update");

  while (gtk_events_pending())
    gtk_main_iteration();

  for (i=0; i<220; i++) {
    sg->min[i] = 64;
    sg->max[i] = 64;
  }
  sg->min[0] = 54;
  sg->max[0] = 74;
  sg->vmin = 0;
  sg->vmax = 0;

  n = 1 << ((int)gtk_range_get_value(GTK_RANGE(sg->impulse_length_scale)));

  sverb_reset(sg->s);

  vmin = vmax = y = sg->s->sverb_process(sg->s, 1.);
  for (j=1; j<n; j++) {
    y = sg->s->sverb_process(sg->s, 0.);
    inc(k);
    if (y > vmax) vmax = y;
    if (y < vmin) vmin = y;
  }
  totmin = vmin;
  totmax = vmax;
  min[0] = vmin;
  max[0] = vmax;

  for (i=1; i<220; i++) {
    vmin = vmax = y = sg->s->sverb_process(sg->s, 0.);
    inc(k);
    for (j=1; j<n; j++) {
      y = sg->s->sverb_process(sg->s, 0.);
      inc(k);
      if (y > vmax) vmax = y;
      if (y < vmin) vmin = y;
    }
    if (vmin < totmin) totmin = vmin;
    if (vmax > totmax) totmax = vmax;
    min[i] = vmin;
    max[i] = vmax;

    sg->min[i] = 54;
    sg->max[i] = 74;
  }

  if (totmax-totmin < 0.1) {
    float n;
    n = (totmax+totmin)/2 - 0.5;
    totmax = (totmax+totmin)/2 + 0.5;
    totmin = n;
  }

  sg->vmin = totmin;
  sg->vmax = totmax;

  /*
   * totmin -> 128
   *      x -> y
   * totmax -> 0
   *
   * (totmin-totmax)/128 = (x-totmax)/y
   * y = (x-totmax) * 128 / (totmin - totmax)
   */
  for (i=0; i<220; i++) {
    sg->min[i] = (min[i] - totmax) * 128 / (totmin - totmax);
    sg->max[i] = (max[i] - totmax) * 128 / (totmin - totmax);
  }

done:
  gtk_button_set_label(GTK_BUTTON(sg->update_impulse_button),
      "Update impulse response");

  sg->update_running = 0;
}

/* return 1 if prime, 0 if not (we should do something faster, no?) */
static int is_prime(int x)
{
  int i;
  for (i=2; i<=sqrt(x)+1; i++)
    if ((x % i) == 0) return 0;
  return 1;
}

static int primify(int d)
{
  if (d <= 3) return d;
  if (!(d&1)) d++;
  while (!is_prime(d)) d+=2;
  return d;
}

/* GUI behavior management functions */

static void yes_event(int *yes, gint id)
{
  if (id == GTK_RESPONSE_YES)
    *yes = 1;
}

static void about(sverb_gui *sg)
{
  char *about =
"sverb " SVERB_VERSION "\n"
"A basic order 15 CFDN reverb.\n"
"It should sound good.\n"
"http://sed.free.fr/sverb\n"
"Author: C%c%cdric Roux <sed@free.fr>\n"
"Dedicated to all the classical guitarists around this planet.\n"
"Public domain.\n"
"Take your freedom.\n"
"(NOTE: this is a beta release, I need you to find good parameters."
" Contact me when you find good values, thank you.)";
  GtkWidget *dialog = gtk_message_dialog_new(GTK_WINDOW(sg->main_win),
        GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_MODAL,
        GTK_MESSAGE_INFO,
        GTK_BUTTONS_OK,
        /* 195 169 is UTF-8 of "" (&eacute;) */
        about, 195, 169);

  g_signal_connect(dialog, "response", G_CALLBACK(gtk_widget_destroy), dialog);
  g_signal_connect(dialog, "destroy", G_CALLBACK(gtk_main_quit), 0);

  gtk_widget_show_all(dialog);

  gtk_main();
}

/* return 1 if yes, 0 if no */
static int yes_no(void *parent, char *mess)
{
  int rep=0;

  GtkWidget *dialog = gtk_message_dialog_new(parent,
        GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_MODAL,
        GTK_MESSAGE_QUESTION,
        GTK_BUTTONS_YES_NO,
        "%s", mess);

  g_signal_connect_swapped(dialog, "response", G_CALLBACK(yes_event), &rep);
  g_signal_connect(dialog, "response", G_CALLBACK(gtk_widget_destroy), dialog);
  g_signal_connect(dialog, "destroy", G_CALLBACK(gtk_main_quit), 0);

  gtk_widget_show_all(dialog);

  gtk_main();

  return rep;
}

static void alert_file(void *parent, char *mess, char *f)
{
  GtkWidget *dialog = gtk_message_dialog_new(parent,
        GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_MODAL,
        GTK_MESSAGE_ERROR,
        GTK_BUTTONS_OK,
        "%s:\n%s", f, mess);

  g_signal_connect(dialog, "response", G_CALLBACK(gtk_widget_destroy), dialog);
  g_signal_connect(dialog, "destroy", G_CALLBACK(gtk_main_quit), 0);

  gtk_widget_show_all(dialog);

  gtk_main();
}

static void alert(void *parent, char *mess)
{
  GtkWidget *dialog = gtk_message_dialog_new(parent,
        GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_MODAL,
        GTK_MESSAGE_ERROR,
        GTK_BUTTONS_OK,
        "%s", mess);

  g_signal_connect(dialog, "response", G_CALLBACK(gtk_widget_destroy), dialog);
  g_signal_connect(dialog, "destroy", G_CALLBACK(gtk_main_quit), 0);

  gtk_widget_show_all(dialog);

  gtk_main();
}

/* file selector stuff */
static int ok;

static GtkWidget *file_selector;

static char file_name[FILENAME_MAX+1];

/* return 1 if file exist, 0 if not, no check done about file type */
static int file_exist(char *s)
{
  struct stat buf;
  return stat(s, &buf) != -1;
}

static void file_event(void *unused, int *check_file)
{
  char *name = (char *)gtk_file_selection_get_filename(
      GTK_FILE_SELECTION(file_selector));

  if (strlen(name) <= 0 || strlen(name) > FILENAME_MAX) {
    alert(file_selector, "bad file name");
    return;
  }

  strcpy(file_name, name);

  if (*check_file) {
    if (file_exist(file_name))
      if (!yes_no(file_selector, "The file already exists. Are you sure?"))
        return;
  }

  ok = 1;
  gtk_widget_destroy(file_selector);
}

static void select_file_destroy_event(void)
{
  gtk_main_quit();
}

/* return 1 if the file name has been selected (and if check_file is 1
 * and the file already exists, the user has said OK to overwrite/whatever
 * the file
 * return 0 if cancel
 */
int select_file(char *lab, sverb_gui *sg, int check_file)
{
  GtkWidget *win;
  GtkWidget *label = gtk_label_new(lab);
  GtkWidget *sep = gtk_hseparator_new();

  win = gtk_file_selection_new(lab);

  gtk_box_pack_start(GTK_BOX(GTK_FILE_SELECTION(win)->main_vbox), label,
      FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(GTK_FILE_SELECTION(win)->main_vbox), sep,
      FALSE, FALSE, 10);
  gtk_box_reorder_child(GTK_BOX(GTK_FILE_SELECTION(win)->main_vbox), label, 0);
  gtk_box_reorder_child(GTK_BOX(GTK_FILE_SELECTION(win)->main_vbox), sep, 1);

  file_selector = win;

  g_signal_connect(GTK_FILE_SELECTION(win)->ok_button, "clicked",
      G_CALLBACK(file_event), &check_file);

#if 0
  g_signal_connect_swapped(GTK_FILE_SELECTION(win)->ok_button, "clicked",
      G_CALLBACK(gtk_widget_destroy), win);
#endif
  g_signal_connect_swapped(GTK_FILE_SELECTION(win)->cancel_button, "clicked",
      G_CALLBACK(gtk_widget_destroy), win);

  g_signal_connect(win, "destroy", G_CALLBACK(select_file_destroy_event), 0);

  /* I don't like modal stuff, but it's more simpler here (I tried to do
   * it with out it, it becomes very very complex, I don't understand all
   * what happens, the program may crash and I don't want that)
   */
  gtk_window_set_modal(GTK_WINDOW(win), TRUE);

  gtk_window_set_transient_for(GTK_WINDOW(win), GTK_WINDOW(sg->main_win));
  gtk_window_set_destroy_with_parent(GTK_WINDOW(win), TRUE);

  gtk_widget_show_all(win);

  ok = 0;

  gtk_main();

  return ok;
}

/* return 1 if user accepts to modify stuff, 0 if no */
int can_modify(sverb_gui *sg)
{
  if (!sg->modified)
    return 1;
  return yes_no(sg->main_win,
      "Are you sure? (your work is not saved, it will be erased)");
}

static void set_total_delay(sverb_gui *sg)
{
  int i;
  int n=0;
  float len;
  char z[128];

  for (i=0; i<15; i++)
    n += sg->s->DELAY[i];

  len = (float)n/(0.15 * 44100);

  sprintf(z, "(%d, %gs)", n, len);

  gtk_label_set_text(GTK_LABEL(sg->total_delay_label), z);
}

/* signal handling functions */

static void do_load(sverb_gui *sg, char *fn)
{
  FILE *f;
  int maj;
  int min;
  int i;
  int n;
  int delay[15];
  float t60_0;
  float t60_pi;
  int integer_precision;
  char v[64];
  int round =
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(sg->prime_toggle));

  if (!can_modify(sg))
    return;

  f = fopen(fn, "r");
  if (!f) {
    alert_file(sg->main_win, "can't open file", fn);
    return;
  }

  if (fscanf(f, "[sverb %d.%d]\n", &maj, &min) != 2)
    goto bad_file;

  if (maj != 0 || min != 9) {
    fclose(f);
    alert_file(sg->main_win, "bad version", fn);
    return;
  }

  for (i=0; i<15; i++) {
    if (fscanf(f, "delay line %d: %d\n", &n, &delay[i]) != 2)
      goto bad_file;
    if (n != i || delay[i] < 1)
      goto bad_file;
  }

  if (fscanf(f, "t60(0): %g\n", &t60_0) != 1)
    goto bad_file;

  if (fscanf(f, "t60(pi): %g\n", &t60_pi) != 1)
    goto bad_file;

  if (fscanf(f, "integer precision: %d\n", &integer_precision) != 1)
    goto bad_file;
  if (integer_precision < 1 || integer_precision > 31)
    goto bad_file;

  fclose(f);

  strcpy(sg->file, file_name);

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sg->prime_toggle), FALSE);

  /* doing it that way is slow. We should set all the parameters
   * then do GUI update stuff, then and only then propagate values
   * to the underlying sverb, with ony one call to something like
   * sverb_init_with_values(delay, t60x, int_precision, etc.)
   * but it's simpler this way and it's not that slow, is it?
   */
  for (i=0; i<15; i++) {
    sprintf(v, "%d", delay[i]);
    gtk_entry_set_text(GTK_ENTRY(sg->delay_entry[i]), v);
    gtk_widget_activate(sg->delay_entry[i]);
  }

  sprintf(v, "%g", t60_0);
  gtk_entry_set_text(GTK_ENTRY(sg->t60_0_entry), v);
  gtk_widget_activate(sg->t60_0_entry);

  sprintf(v, "%g", t60_pi);
  gtk_entry_set_text(GTK_ENTRY(sg->t60_pi_entry), v);
  gtk_widget_activate(sg->t60_pi_entry);

  gtk_range_set_value(GTK_RANGE(sg->int_precision_scale), integer_precision);

  sg->modified = 0;

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sg->prime_toggle), round);

  return;

bad_file:
  fclose(f);
  alert_file(sg->main_win, "bad file", fn);
}

static void toggle_help_callback(void *unused, sverb_gui *sg)
{
  if (sg->help_window_shown)
    gtk_widget_hide_all(sg->help_window);
  else
    gtk_widget_show_all(sg->help_window);
  sg->help_window_shown = 1 - sg->help_window_shown;
}

static gboolean delete_help_callback(void *u1, void *u2, sverb_gui *sg)
{
  toggle_help_callback(0, sg);
  return TRUE;
}

static void load_callback(GtkMenuItem *m, sverb_gui *sg)
{
  if (select_file("Load", sg, 0) == 0)
    return;
  do_load(sg, file_name);
}

static void do_save(sverb_gui *sg, char *fn)
{
  FILE *f = 0;
  int i;

  f = fopen(fn, "w");
  if (!f) {
    alert_file(sg->main_win, "can't save", fn);
    return;
  }

  /* don't use SVERB_VERSION here, the version of the file format is
   * currently 0.9 and should not change if no change in the save
   * format is done
   */
  if (fprintf(f, "[sverb 0.9]\n") == -1)
    goto save_error;

  for (i=0; i<15; i++)
    if (fprintf(f, "delay line %d: %d\n", i, sg->s->DELAY[i]) == -1)
      goto save_error;

  if (fprintf(f, "t60(0): %g\n", sg->s->t60_0) == -1)
    goto save_error;

  if (fprintf(f, "t60(pi): %g\n", sg->s->t60_pi) == -1)
    goto save_error;

  if (fprintf(f, "integer precision: %d\n", sg->s->integer_precision) == -1)
    goto save_error;

  if (fclose(f) == -1) {
    f = 0;
    goto save_error;
  }

  strcpy(sg->file, fn);
  sg->modified = 0;

  return;

save_error:
  fclose(f);
  alert_file(sg->main_win, "error while saving (no space on disk?)", fn);
}

static void saveas_callback(GtkMenuItem *m, sverb_gui *sg)
{
  if (select_file(m ? "Save as" : "Save", sg, 1) == 0)
    return;
  do_save(sg, file_name);
}

static void save_callback(GtkMenuItem *m, sverb_gui *sg)
{
  if (!sg->file[0])
    return saveas_callback(0, sg);
  do_save(sg, sg->file);
}

/* forward declaration */
static void update_impulse_callback(GtkButton *, sverb_gui *);

static void update_menu_callback(GtkMenuItem *m, sverb_gui *sg)
{
  update_impulse_callback(GTK_BUTTON(sg->update_impulse_button), sg);
}

static void play_menu_callback(GtkMenuItem *m, sverb_gui *sg)
{
  impulse_play_callback(GTK_BUTTON(sg->impulse_play_button), sg);
}

static void quit_callback(GtkMenuItem *m, sverb_gui *sg)
{
  if (sg->modified)
    if (!yes_no(sg->main_win, "Are you sure you want to quit?"))
      return;
  gtk_main_quit();
}

static void help_callback(GtkMenuItem *m, sverb_gui *sg)
{
  toggle_help_callback(0, sg);
}

static void about_callback(GtkMenuItem *m, sverb_gui *sg)
{
  about(sg);
}

static char *names[] = {
      "tiny",
      "small 1",
      "small 2",
      "medium 1",
      "medium 2",
      "average 1",
      "average 2",
      "average 3",
      "middle",
      "large 1",
      "large 2",
      "big 1",
      "big 2",
      "huge",
      0
    };
static void preset_callback(GtkButton *preset, sverb_gui *sg)
{
  static int delays[][15] = {
    /* tiny */
    { 42, 29, 26, 23, 21, 19, 18, 17, 16, 15, 14, 13, 11, 9, 7 },
    /* small 1 */
    {1, 2, 109, 107, 5, 6, 7, 19, 29, 10, 137, 12, 13, 191, 15},
    /* small 2 */
    {15, 14, 113, 12, 11, 10, 59, 8, 7, 233, 97, 4, 29, 2, 89},
    /* medium 1 */
    {907, 777, 631, 313, 211, 137, 107, 99, 69, 43, 21, 13, 11, 9, 3},
    /* medium 2 */
    {197, 179, 113, 163, 199, 29, 379, 137, 103, 457, 389, 23, 97, 653, 223},
    /* normal 1 */
    {2207, 223, 307, 59, 647, 149, 359, 83, 127, 613, 593, 61, 337, 53, 877},
    /* normal 2 */
    {157, 17, 113, 313, 613, 1823, 521, 1523, 107, 73, 61, 853, 269, 281, 31},
    /* normal 3 */
    {733, 317, 367, 541, 401, 139, 379, 607, 233, 727, 467, 263, 163, 929,
     443},
    /* middle */
    {3407, 2551, 2053, 1867, 1009, 641, 367, 313, 269, 113, 53, 43, 37,
     29, 19},
    /* large 1 */
    {1627, 401, 577, 6563, 1319, 3691, 1259, 2203, 3511, 1187, 673, 809,
     4211, 4889, 163},
    /* large 2 */
    {2647, 1373, 1993, 2909, 2111, 1901, 2423, 2789, 1439, 2459, 1931,
     2179, 1907, 2633, 2437},
    /* big 1 */
    {2011, 1901, 3527, 7283, 4099, 4463, 3821, 5189, 3581, 8419, 5323,
     8707, 3407, 3727, 691},
    /* big 2 */
    {4903, 4243, 3733, 5557, 4127, 4441, 5419, 6211, 3299, 4229, 3631,
     3727, 3907, 4079, 4673},
    /* huge */
    {2153, 7411, 487, 10559, 15233, 8209, 2069, 8387, 10627, 18041,
     14447, 20743, 2063, 8951, 3943}
  };
  static float t60_0s[] = {
    0.04,
    0.1,
    0.1,
    0.5,
    0.5,
    1.,
    1.,
    1.,
    2.,
    5.,
    5.,
    10.,
    10.,
    20.
  };
  static float t60_pis[] = {
    0.02,
    0.05,
    0.05,
    0.25,
    0.25,
    0.5,
    0.5,
    0.5,
    1.,
    2.5,
    2.5,
    5.,
    5.,
    10.
  };

  char n[64];
  const char *s =
      gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(sg->combo_preset)->entry));
  int *delay = 0;
  float t60_0;
  float t60_pi;
  int i;

  if (!can_modify(sg))
    return;

  t60_0 = 0;  /* for no gcc warnings */
  t60_pi = 0; /* for no gcc warnings */

  for (i=0; names[i]; i++) {
    if (!strcmp(s, names[i])) {
      delay = delays[i];
      t60_0 = t60_0s[i];
      t60_pi = t60_pis[i];
      break;
    }
  }

  /* just in case (it should not happen) */
  if (!delay) {
    delay = delays[0];
    t60_0 = t60_0s[0];
    t60_pi = t60_pis[0];
  }

  for (i = 0; i < 15; i++) {
    sverb_set_delay(sg->s, i, delay[i]);
    sprintf(n, "%d", sg->s->DELAY[i]);
    gtk_entry_set_text(GTK_ENTRY(sg->delay_entry[i]), n);
    gtk_label_set_text(GTK_LABEL(sg->delay_label[i]), n);
  }

  sverb_set_t60(sg->s, t60_0, t60_pi);
  sprintf(n, "%g", t60_0);
  gtk_entry_set_text(GTK_ENTRY(sg->t60_0_entry), n);
  sprintf(n, "t60(0): %gs", t60_0);
  gtk_label_set_text(GTK_LABEL(sg->t60_0_label), n);
  sprintf(n, "%g", t60_pi);
  gtk_entry_set_text(GTK_ENTRY(sg->t60_pi_entry), n);
  sprintf(n, "t60(PI): %gs", t60_pi);
  gtk_label_set_text(GTK_LABEL(sg->t60_pi_label), n);

  set_total_delay(sg);

  sg->modified = 0;
}

static void delay_entry_callback(GtkWidget *e, sverb_gui *sg)
{
  int old;
  int d;
  const char *s;
  char n[16];
  int i;

  for (i=0; i<15; i++)
    if (sg->delay_entry[i] == e) break;
  s = gtk_entry_get_text(GTK_ENTRY(sg->delay_entry[i]));
  sscanf(gtk_label_get_text(GTK_LABEL(sg->delay_label[i])), "%d", &old);
  s = gtk_entry_get_text(GTK_ENTRY(sg->delay_entry[i]));
  if (sscanf(s, "%d", &d) != 1 || d < 1) {
    gtk_entry_set_text(GTK_ENTRY(sg->delay_entry[i]),
        gtk_label_get_text(GTK_LABEL(sg->delay_label[i])));
    return;
  }
  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(sg->prime_toggle))
      == TRUE)
    d = primify(d);
  sprintf(n, "%d", d);
  gtk_entry_set_text(GTK_ENTRY(sg->delay_entry[i]), n);
  gtk_label_set_text(GTK_LABEL(sg->delay_label[i]), n);
  sg->modified = 1;
  if (sverb_set_delay(sg->s, i, d) == -1) {
    alert(sg->main_win, "Error, probably no more memory available");
    sprintf(n, "%d", old);
    gtk_entry_set_text(GTK_ENTRY(sg->delay_entry[i]), n);
    gtk_label_set_text(GTK_LABEL(sg->delay_label[i]), n);
    sg->modified = 0;
    return;
  }
  set_total_delay(sg);
}

static void t60_0_callback(GtkEntry *e, sverb_gui *sg)
{
  float t60_0, t60_pi;
  float new;
  char n[64];

  sscanf(gtk_label_get_text(GTK_LABEL(sg->t60_0_label)),
      "t60(0): %gs", &t60_0);
  sscanf(gtk_label_get_text(GTK_LABEL(sg->t60_pi_label)),
      "t60(PI): %gs", &t60_pi);
  if (sscanf(gtk_entry_get_text(GTK_ENTRY(sg->t60_0_entry)),
      "%g", &new) != 1) {
    sprintf(n, "%g", t60_0);
    gtk_entry_set_text(GTK_ENTRY(sg->t60_0_entry), n);
    return;
  }
  sverb_set_t60(sg->s, new, t60_pi);
  sprintf(n, "%g", new);
  gtk_entry_set_text(GTK_ENTRY(sg->t60_0_entry), n);
  sprintf(n, "t60(0): %gs", new);
  gtk_label_set_text(GTK_LABEL(sg->t60_0_label), n);
  sg->modified = 1;
}

static void t60_pi_callback(GtkEntry *e, sverb_gui *sg)
{
  float t60_0, t60_pi;
  float new;
  char n[64];

  sscanf(gtk_label_get_text(GTK_LABEL(sg->t60_0_label)),
      "t60(0): %gs", &t60_0);
  sscanf(gtk_label_get_text(GTK_LABEL(sg->t60_pi_label)),
      "t60(PI): %gs", &t60_pi);
  if (sscanf(gtk_entry_get_text(GTK_ENTRY(sg->t60_pi_entry)),
      "%g", &new) != 1) {
    sprintf(n, "%g", t60_pi);
    gtk_entry_set_text(GTK_ENTRY(sg->t60_pi_entry), n);
    return;
  }
  sverb_set_t60(sg->s, t60_0, new);
  sprintf(n, "%g", new);
  gtk_entry_set_text(GTK_ENTRY(sg->t60_pi_entry), n);
  sprintf(n, "t60(PI): %gs", new);
  gtk_label_set_text(GTK_LABEL(sg->t60_pi_label), n);
  sg->modified = 1;
}

static void bypass_t60_callback(GtkToggleButton *b, sverb_gui *sg)
{
  sverb_use_t60(sg->s, gtk_toggle_button_get_active(
      GTK_TOGGLE_BUTTON(sg->bypass_t60_toggle)) == FALSE);
}

static void precision_callback(GtkToggleButton *b, sverb_gui *sg)
{
  int type;

  if (gtk_toggle_button_get_active(b) == FALSE)
    return;

#ifdef ASM_IA32
  if (b == (GtkToggleButton *)sg->asm_button)
    type = SVERB_ASM;
  else
#endif
  if (b == (GtkToggleButton *)sg->float_button)
    type = SVERB_FLOAT;
  else
    type = SVERB_INT;

  sverb_set_type(sg->s, type);
}

static gchar *integer_precision_format_callback(GtkScale *s, gdouble v)
{
  /* this malloc looks really bad, but hey, gtk... */
  return g_strdup_printf("%d bit%s", (int)v, v > 1 ? "s" : "");
}

static void integer_precision_callback(GtkRange *r, sverb_gui *sg)
{
  sverb_set_integer_precision(sg->s, gtk_range_get_value(r));
}

static void impulse_expose_callback(GtkWidget *w, GdkEventExpose *e,
    sverb_gui *sg)
{
  int i;
  GdkSegment segs[220];
  char n[64];

  for (i=0; i<220; i++) {
    segs[i].x1 = i;
    segs[i].x2 = i;
    segs[i].y1 = sg->min[i];
    segs[i].y2 = sg->max[i];
  }

  gdk_draw_segments(w->window, w->style->black_gc, segs, 220);

  sprintf(n, "%g", sg->vmax);
  gtk_label_set_text(GTK_LABEL(sg->impulse_max), n);
  sprintf(n, "%g", sg->vmin);
  gtk_label_set_text(GTK_LABEL(sg->impulse_min), n);
}

/* forward declaration */
static void impulse_play_callback(GtkButton *, sverb_gui *);

static void update_impulse_callback(GtkButton *b, sverb_gui *sg)
{
  if (sg->update_running) {
    sg->aborted = 1;
    return;
  }

  /* if playing, we first stop the playing */
  if (sg->playing) {
    impulse_play_callback(GTK_BUTTON(sg->impulse_play_button), sg);
  }

  calculate_response(sg);
  gtk_widget_queue_draw(sg->impulse_drawing_area);
}

static gchar *impulse_length_format_callback(GtkScale *s, gdouble v)
{
  /* this malloc look really bad, but hey, gtk... */
  int n = 1 << (int)v;
  if (n == 1)
    return g_strdup_printf("0.005s");
  return g_strdup_printf("%.2fs", (float)n/44100. * 220.);
}

static void impulse_play_callback(GtkButton *b, sverb_gui *sg)
{
  if (sg->playing) {
    sg->playing = 0;
    close_sound();
    gtk_input_remove(sg->audio_callback);
    gtk_button_set_label(GTK_BUTTON(sg->impulse_play_button), " Play ");
    return;
  }

  /* if updating, stop update */
  if (sg->update_running)
    update_impulse_callback(GTK_BUTTON(sg->update_impulse_button), sg);

  if (!init_sound((char *)gtk_entry_get_text(GTK_ENTRY(sg->oss_output_entry)),
      44100, 16, 0, 12, PLAY_SOUND)) {
    alert(sg->main_win, "Error with audio. Try another device. "
        "Maybe you don't have OSS, or maybe some other program already "
        "uses the soundcard. If you have KDE or Gnome, these stuffs tend "
        "to take over the audio device.");
    return;
  }
  sverb_reset(sg->s);
  sg->audio_callback = gtk_input_add_full(get_sound_fd(), GDK_INPUT_WRITE,
      (GdkInputFunction)audio_ready, 0, sg, 0);
  sg->playing = 3;
  gtk_button_set_label(GTK_BUTTON(sg->impulse_play_button), " Stop ");
  sscanf(gtk_label_get_text(GTK_LABEL(sg->input_scale_label)),
      "Input scale: %g", &sg->iscale);
  sscanf(gtk_label_get_text(GTK_LABEL(sg->output_scale_label)),
      "Output scale: %g", &sg->oscale);
}

static void input_scale_callback(GtkEntry *e, sverb_gui *sg)
{
  char n[64];
  float scale;
  float ns;
  sscanf(gtk_label_get_text(GTK_LABEL(sg->input_scale_label)),
      "Input scale: %g", &scale);
  if (sscanf(gtk_entry_get_text(e), "%g", &ns) != 1) {
    sprintf(n, "%g", scale);
    gtk_entry_set_text(e, n);
    sprintf(n, "Input scale: %g", scale);
    gtk_label_set_text(GTK_LABEL(sg->input_scale_label), n);
    return;
  }
  sprintf(n, "%g", ns);
  gtk_entry_set_text(e, n);
  sprintf(n, "Input scale: %g", ns);
  gtk_label_set_text(GTK_LABEL(sg->input_scale_label), n);
}

static void output_scale_callback(GtkEntry *e, sverb_gui *sg)
{
  char n[64];
  float scale;
  float ns;
  sscanf(gtk_label_get_text(GTK_LABEL(sg->output_scale_label)),
      "Output scale: %g", &scale);
  if (sscanf(gtk_entry_get_text(e), "%g", &ns) != 1) {
    sprintf(n, "%g", scale);
    gtk_entry_set_text(e, n);
    sprintf(n, "Output scale: %g", scale);
    gtk_label_set_text(GTK_LABEL(sg->output_scale_label), n);
    return;
  }
  sprintf(n, "%g", ns);
  gtk_entry_set_text(e, n);
  sprintf(n, "Output scale: %g", ns);
  gtk_label_set_text(GTK_LABEL(sg->output_scale_label), n);
}

/* init */
void sverb_gui_init(sverb_gui *sg, sverb *s)
{
  int i;
  GtkAccelGroup *main_accel_group;
  GtkWidget *main_vbox_main;
  GtkWidget *main_vbox;
  GtkWidget *main_hbox;
  GtkWidget *main_menubar;
  GtkWidget *main_menubar_file;
  GtkWidget *main_menubar_help;
  GtkWidget *menu_file;
  GtkWidget *menu_file_load;
  GtkWidget *menu_file_save;
  GtkWidget *menu_file_saveas;
  GtkWidget *menu_file_update_impulse;
  GtkWidget *menu_file_play;
  GtkWidget *menu_file_quit;
  GtkWidget *menu_help;
  GtkWidget *menu_help_help;
  GtkWidget *menu_help_about;
  GtkWidget *hbox;
  GtkWidget *vbox;
  GtkWidget *preset_button;
  GList *list_preset = 0;
  GList *new;
  char n[64];

  sg->modified = 0;
  sg->aborted = 0;
  sg->update_running = 0;

  sg->playing = 0;

  sg->file[0] = 0;

  sg->s = s;

  /* main window */
  sg->main_win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  {
    /* let's force its minimal size to be 1x1, I don't like to let
     * the user with a too big window if she wants to resize it smaller
     */
    GdkGeometry g;
    g.min_width = g.min_height = 1;
    gtk_window_set_geometry_hints(GTK_WINDOW(sg->main_win),
                                  sg->main_win, &g, GDK_HINT_MIN_SIZE);
  }
  //gtk_widget_set_size_request(sg->main_win, 700, 500);
  gtk_window_set_title(GTK_WINDOW(sg->main_win), "sverb " SVERB_VERSION);

  main_vbox_main = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(sg->main_win), main_vbox_main);
  main_hbox = gtk_hbox_new(FALSE, 0);
  main_vbox = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(main_hbox), main_vbox, FALSE, FALSE, 10);

  /* menu: FILE (load / save / quit), HELP (about) */
  main_menubar = gtk_menu_bar_new();
  gtk_box_pack_start(GTK_BOX(main_vbox_main), main_menubar, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(main_vbox_main), main_hbox, FALSE, FALSE, 0);
  main_menubar_file = gtk_menu_item_new_with_mnemonic("_File");
  main_menubar_help = gtk_menu_item_new_with_mnemonic("_Help");
  gtk_menu_shell_append(GTK_MENU_SHELL(main_menubar), main_menubar_file);
  gtk_menu_shell_append(GTK_MENU_SHELL(main_menubar), main_menubar_help);

  menu_file = gtk_menu_new();
  menu_file_load = gtk_menu_item_new_with_mnemonic("_Load");
  menu_file_save = gtk_menu_item_new_with_mnemonic("_Save");
  menu_file_saveas = gtk_menu_item_new_with_mnemonic("Save _as");
  menu_file_update_impulse =
      gtk_menu_item_new_with_mnemonic("_Update Impulse");
  menu_file_play = gtk_menu_item_new_with_mnemonic("_Play / Stop");
  menu_file_quit = gtk_menu_item_new_with_mnemonic("_Quit");
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(main_menubar_file), menu_file);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_file), menu_file_load);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_file), menu_file_save);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_file), menu_file_saveas);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_file),
      gtk_separator_menu_item_new());
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_file), menu_file_update_impulse);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_file), menu_file_play);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_file),
      gtk_separator_menu_item_new());
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_file), menu_file_quit);

  menu_help = gtk_menu_new();
  menu_help_help = gtk_menu_item_new_with_mnemonic("_Help");
  menu_help_about = gtk_menu_item_new_with_mnemonic("_About");
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(main_menubar_help), menu_help);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_help), menu_help_help);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_help),
      gtk_separator_menu_item_new());
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_help), menu_help_about);

  /* accel group for the menu */
  main_accel_group = gtk_accel_group_new();

  gtk_accel_map_add_entry("<sverb>/File/Load", GDK_l, GDK_CONTROL_MASK);
  gtk_accel_map_add_entry("<sverb>/File/Save", GDK_s, GDK_CONTROL_MASK);
  gtk_accel_map_add_entry("<sverb>/File/Update", GDK_u, GDK_CONTROL_MASK);
  gtk_accel_map_add_entry("<sverb>/File/Play", GDK_p, GDK_CONTROL_MASK);
  gtk_accel_map_add_entry("<sverb>/File/Quit", GDK_q, GDK_CONTROL_MASK);
  gtk_accel_map_add_entry("<sverb>/Help/Help", GDK_h, GDK_CONTROL_MASK);
  gtk_accel_map_add_entry("<sverb>/Help/About", GDK_a, GDK_CONTROL_MASK);

  gtk_window_add_accel_group(GTK_WINDOW(sg->main_win), main_accel_group);
  gtk_menu_set_accel_group(GTK_MENU(menu_file), main_accel_group);
  gtk_menu_set_accel_group(GTK_MENU(menu_help), main_accel_group);

  gtk_menu_item_set_accel_path(GTK_MENU_ITEM(menu_file_load),
      "<sverb>/File/Load");
  gtk_menu_item_set_accel_path(GTK_MENU_ITEM(menu_file_save),
      "<sverb>/File/Save");
  gtk_menu_item_set_accel_path(GTK_MENU_ITEM(menu_file_update_impulse),
      "<sverb>/File/Update");
  gtk_menu_item_set_accel_path(GTK_MENU_ITEM(menu_file_play),
      "<sverb>/File/Play");
  gtk_menu_item_set_accel_path(GTK_MENU_ITEM(menu_file_quit),
      "<sverb>/File/Quit");
  gtk_menu_item_set_accel_path(GTK_MENU_ITEM(menu_help_help),
      "<sverb>/Help/Help");
  gtk_menu_item_set_accel_path(GTK_MENU_ITEM(menu_help_about),
      "<sverb>/Help/About");

  /* preset */
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(main_vbox), hbox, FALSE, FALSE, 10);
  gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("Preset"), FALSE, FALSE, 10);
  sg->combo_preset = gtk_combo_new();
  gtk_box_pack_start(GTK_BOX(hbox), sg->combo_preset, FALSE, FALSE, 0);
  gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(sg->combo_preset)->entry), FALSE);
  {
    for (i=0; names[i]; i++) {
      new = g_list_append(list_preset, names[i]);
      if (!new) goto list_error;
      list_preset = new;
    }
  }
  gtk_combo_set_popdown_strings(GTK_COMBO(sg->combo_preset), list_preset);
  g_list_free(list_preset);
  preset_button = gtk_button_new_with_label("Set");
  gtk_box_pack_start(GTK_BOX(hbox), preset_button, FALSE, FALSE, 0);

  /* delays */
  {
    int i;

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(main_vbox), hbox, FALSE, FALSE, 10);
    gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("Delay lines"),
        FALSE, FALSE, 10);
    sg->total_delay_label = gtk_label_new("0");
    gtk_box_pack_start(GTK_BOX(hbox), sg->total_delay_label, FALSE, FALSE, 10);

    for (i=0; i<15; i++) {
      if (!(i % 5)) {
        hbox = gtk_hbox_new(FALSE, 3);
        gtk_box_pack_start(GTK_BOX(main_vbox), hbox, FALSE, FALSE, 5);
      }
      vbox = gtk_vbox_new(FALSE, 2);
      gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE,
          i%5 == 0 ? 10 : (i%5)%2 ? 0 : 10);
      sprintf(n, "%d", s->DELAY[i]);
      sg->delay_label[i] = gtk_label_new(n);
      gtk_box_pack_start(GTK_BOX(vbox), sg->delay_label[i], FALSE, FALSE, 0);
      sg->delay_entry[i] = gtk_entry_new();
      gtk_widget_set_size_request(sg->delay_entry[i], 40, -1);
      gtk_entry_set_text(GTK_ENTRY(sg->delay_entry[i]), n);
      gtk_box_pack_start(GTK_BOX(vbox), sg->delay_entry[i], FALSE, FALSE, 0);
    }
  }

  /* round to up prime */
  hbox = gtk_hbox_new(FALSE, 5);
  sg->prime_toggle =
      gtk_check_button_new_with_mnemonic("_Round delay to upper prime");
  gtk_box_pack_start(GTK_BOX(main_vbox), hbox, FALSE, FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), sg->prime_toggle, FALSE, FALSE, 10);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sg->prime_toggle), TRUE);

  /* t60_0 t60_pi use t60 */
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(main_vbox), hbox, FALSE, FALSE, 5);
  vbox = gtk_vbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 10);
  sprintf(n, "t60(0): %gs", s->t60_0);
  sg->t60_0_label = gtk_label_new(n);
  gtk_box_pack_start(GTK_BOX(vbox), sg->t60_0_label, FALSE, FALSE, 0);
  sg->t60_0_entry = gtk_entry_new();
  gtk_widget_set_size_request(sg->t60_0_entry, 80, -1);
  sprintf(n, "%g", s->t60_0);
  gtk_entry_set_text(GTK_ENTRY(sg->t60_0_entry), n);
  gtk_box_pack_start(GTK_BOX(vbox), sg->t60_0_entry, FALSE, FALSE, 0);
  vbox = gtk_vbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
  sprintf(n, "t60(PI): %gs", s->t60_pi);
  sg->t60_pi_label = gtk_label_new(n);
  gtk_box_pack_start(GTK_BOX(vbox), sg->t60_pi_label, FALSE, FALSE, 0);
  sg->t60_pi_entry = gtk_entry_new();
  gtk_widget_set_size_request(sg->t60_pi_entry, 80, -1);
  sprintf(n, "%g", s->t60_pi);
  gtk_entry_set_text(GTK_ENTRY(sg->t60_pi_entry), n);
  gtk_box_pack_start(GTK_BOX(vbox), sg->t60_pi_entry, FALSE, FALSE, 0);
  hbox = gtk_hbox_new(FALSE, 5);
  sg->bypass_t60_toggle = gtk_check_button_new_with_mnemonic(
      "_Bypass t60 (for a lossless reverb)");
  gtk_box_pack_start(GTK_BOX(main_vbox), hbox, FALSE, FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), sg->bypass_t60_toggle, FALSE, FALSE, 10);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sg->bypass_t60_toggle),
      FALSE);

  /* float / int / asm */
  hbox = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(main_vbox), hbox, FALSE, FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("Precision"),
      FALSE, FALSE, 10);
  sg->float_button = gtk_radio_button_new_with_label(0, "Float");
  gtk_box_pack_start(GTK_BOX(hbox), sg->float_button, FALSE, FALSE, 0);
  sg->int_button = gtk_radio_button_new_with_label_from_widget(
      GTK_RADIO_BUTTON(sg->float_button), "Integer");
  gtk_box_pack_start(GTK_BOX(hbox), sg->int_button, FALSE, FALSE, 0);
#ifdef ASM_IA32
  sg->asm_button = gtk_radio_button_new_with_label_from_widget(
      GTK_RADIO_BUTTON(sg->int_button), "Asm");
  gtk_box_pack_start(GTK_BOX(hbox), sg->asm_button, FALSE, FALSE, 0);
#endif

  /* int precision (nb bits) */
  hbox = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(main_vbox), hbox, FALSE, FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("Integer precision"),
      FALSE, FALSE, 10);
  {
    GtkAdjustment *adj = (GtkAdjustment *)
        gtk_adjustment_new(14, 1, 31, 1, 10, -1);
    sg->int_precision_scale = gtk_hscale_new(adj);
    gtk_scale_set_digits(GTK_SCALE(sg->int_precision_scale), 0);
    gtk_widget_set_size_request(sg->int_precision_scale, 100, -1);
  }
  gtk_box_pack_start(GTK_BOX(hbox), sg->int_precision_scale, FALSE, FALSE, 10);

  /* separator */
  gtk_box_pack_start(GTK_BOX(main_hbox),
      gtk_vseparator_new(), FALSE, FALSE, 10);

  main_vbox = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(main_hbox), main_vbox, FALSE, FALSE, 10);

  /* impulse response */
  sg->impulse_min = gtk_label_new("0");
  sg->impulse_max = gtk_label_new("0");
  hbox = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(main_vbox),hbox, FALSE, FALSE, 10);
  gtk_box_pack_start(GTK_BOX(hbox),
      gtk_label_new("Impulse response"), FALSE, FALSE, 0);
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(main_vbox),hbox, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), sg->impulse_max, FALSE, FALSE, 0);
  sg->impulse_drawing_area = gtk_drawing_area_new();
  gtk_widget_modify_bg(sg->impulse_drawing_area, GTK_STATE_NORMAL,
      &sg->impulse_drawing_area->style->white);
  gtk_widget_set_size_request(sg->impulse_drawing_area, 220, 129);
  hbox = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(main_vbox), hbox, FALSE, FALSE, 10);
  gtk_box_pack_start(GTK_BOX(hbox), sg->impulse_drawing_area,
      FALSE, FALSE, 0);
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(main_vbox),hbox, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), sg->impulse_min, FALSE, FALSE, 0);

  /* update IR, stop update */
  sg->update_impulse_button =
      gtk_button_new_with_label("Update Impulse response");
  hbox = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(main_vbox), hbox, FALSE, FALSE, 10);
  gtk_box_pack_start(GTK_BOX(hbox), sg->update_impulse_button,
      FALSE, FALSE, 0);

  /* 0.1s, 1s, 10s, 100s, 1000s */
  hbox = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(main_vbox), hbox, FALSE, FALSE, 5);
  {
    GtkAdjustment *adj = (GtkAdjustment *)
        gtk_adjustment_new(3, 0, 19, 1, 5, -1);
    sg->impulse_length_scale = gtk_hscale_new(adj);
    gtk_scale_set_digits(GTK_SCALE(sg->impulse_length_scale), 0);
    gtk_widget_set_size_request(sg->impulse_length_scale, 200, -1);
  }
  gtk_box_pack_start(GTK_BOX(hbox), sg->impulse_length_scale, FALSE, FALSE, 0);

  /* play IR / output device */
  hbox = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(main_vbox), hbox, FALSE, FALSE, 5);
  sg->impulse_play_button = gtk_button_new_with_label(" Play ");
  gtk_box_pack_start(GTK_BOX(hbox), sg->impulse_play_button, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("OSS device"),
      FALSE, FALSE, 0);
  sg->oss_output_entry = gtk_entry_new();
  gtk_widget_set_size_request(sg->oss_output_entry, 80, -1);
  gtk_entry_set_text(GTK_ENTRY(sg->oss_output_entry), "/dev/dsp");
  gtk_box_pack_start(GTK_BOX(hbox), sg->oss_output_entry, FALSE, FALSE, 0);

  /* input scale, output scale */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(main_vbox), hbox, FALSE, FALSE, 10);
  vbox = gtk_vbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
  sprintf(n, "Input scale: %g", (float)20000.1);
  sg->input_scale_label = gtk_label_new(n);
  gtk_box_pack_start(GTK_BOX(vbox), sg->input_scale_label, FALSE, FALSE, 0);
  sg->input_scale_entry = gtk_entry_new();
  gtk_widget_set_size_request(sg->input_scale_entry, 80, -1);
  sprintf(n, "%g", (float)20000.1);
  gtk_entry_set_text(GTK_ENTRY(sg->input_scale_entry), n);
  gtk_box_pack_start(GTK_BOX(vbox), sg->input_scale_entry, FALSE, FALSE, 0);
  vbox = gtk_vbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 20);
  sprintf(n, "Output scale: %g", (float)1.001);
  sg->output_scale_label = gtk_label_new(n);
  gtk_box_pack_start(GTK_BOX(vbox), sg->output_scale_label, FALSE, FALSE, 0);
  sg->output_scale_entry = gtk_entry_new();
  gtk_widget_set_size_request(sg->output_scale_entry, 80, -1);
  sprintf(n, "%g", (float)1.001);
  gtk_entry_set_text(GTK_ENTRY(sg->output_scale_entry), n);
  gtk_box_pack_start(GTK_BOX(vbox), sg->output_scale_entry, FALSE, FALSE, 0);

  /* let's create the help window */
  /* it is juste hidden/shown on user's request, maybe not very efficient? */
{
  GtkWidget *help_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  GtkWidget *vbox = gtk_vbox_new(FALSE, 5);
  GtkWidget *hbox = gtk_hbox_new(FALSE, 5);
  GtkWidget *close_button = gtk_button_new_with_mnemonic("_Close");
  GtkWidget *scroll_win = gtk_scrolled_window_new(NULL, NULL);
  GtkTextIter iter;
  GtkTextTagTable *table = gtk_text_tag_table_new();
  GtkTextBuffer *buffer = gtk_text_buffer_new(table);
  GtkWidget *t = gtk_text_view_new_with_buffer(buffer);
  {
    /* let's force minimal size of help_window to be 1x1, I don't like to let
     * the user with a too big window if she wants to resize it smaller
     */
    GdkGeometry g;
    g.min_width = g.min_height = 1;
    gtk_window_set_geometry_hints(GTK_WINDOW(help_window),
                                  help_window, &g, GDK_HINT_MIN_SIZE);
  }
  gtk_window_set_title(GTK_WINDOW(help_window), "sverb help");
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll_win),
      GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_container_add(GTK_CONTAINER(help_window), vbox);
  gtk_box_pack_start(GTK_BOX(vbox), scroll_win, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), close_button, FALSE, FALSE, 5);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(t), GTK_WRAP_WORD);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(t), FALSE);
  gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(t), FALSE);
  gtk_text_view_set_pixels_above_lines(GTK_TEXT_VIEW(t), 5);
  gtk_text_view_set_pixels_below_lines(GTK_TEXT_VIEW(t), 5);
  gtk_text_view_set_left_margin(GTK_TEXT_VIEW(t), 10);
  gtk_text_view_set_right_margin(GTK_TEXT_VIEW(t), 10);
  /* not supported by the gtk on my host... WHY TO PROVIDE THE FILL PARAMETER
   * THEN!!??
  gtk_text_view_set_justification(GTK_TEXT_VIEW(t), GTK_JUSTIFY_FILL);
   */
  gtk_widget_set_size_request(t, 400, 300);
  gtk_container_add(GTK_CONTAINER(scroll_win), t);
  gtk_text_buffer_create_tag(buffer, "b",
      "weight", PANGO_WEIGHT_BOLD,
      0);
  gtk_text_buffer_create_tag(buffer, "p",
      "size", 14 * PANGO_SCALE,
      //"pixels_above_lines", 20,
      0);
  gtk_text_buffer_create_tag(buffer, "h1",
      "justification", GTK_JUSTIFY_CENTER,
      "weight", PANGO_WEIGHT_BOLD,
      "size", 20 * PANGO_SCALE,
      0);
  gtk_text_buffer_create_tag(buffer, "h2",
      "weight", PANGO_WEIGHT_BOLD,
      "size", 18 * PANGO_SCALE,
      0);
  gtk_text_buffer_create_tag(buffer, "h3",
      "weight", PANGO_WEIGHT_BOLD,
      "size", 16 * PANGO_SCALE,
      0);
  gtk_text_buffer_get_iter_at_offset (buffer, &iter, 0);
#include "help_gui.h"
  sg->help_window = help_window;
  sg->help_window_shown = 0;
  g_signal_connect(close_button, "clicked",
      G_CALLBACK(toggle_help_callback), sg);
  g_signal_connect(help_window, "delete-event",
      G_CALLBACK(delete_help_callback), sg);
}

  /* signals */
  g_signal_connect(menu_file_load, "activate",
      G_CALLBACK(load_callback), sg);
  g_signal_connect(menu_file_save, "activate",
      G_CALLBACK(save_callback), sg);
  g_signal_connect(menu_file_saveas, "activate",
      G_CALLBACK(saveas_callback), sg);
  g_signal_connect(menu_file_update_impulse, "activate",
      G_CALLBACK(update_menu_callback), sg);
  g_signal_connect(menu_file_play, "activate",
      G_CALLBACK(play_menu_callback), sg);
  g_signal_connect(menu_file_quit, "activate",
      G_CALLBACK(quit_callback), sg);
  g_signal_connect(menu_help_help, "activate",
      G_CALLBACK(help_callback), sg);
  g_signal_connect(menu_help_about, "activate",
      G_CALLBACK(about_callback), sg);
  g_signal_connect(preset_button, "clicked",
      G_CALLBACK(preset_callback), sg);
  for (i=0; i<15; i++)
    g_signal_connect(sg->delay_entry[i], "activate",
        G_CALLBACK(delay_entry_callback), sg);
#if 0
  g_signal_connect(sg->prime_toggle, "toggled",
      G_CALLBACK(prime_callback), sg);
#endif
  g_signal_connect(sg->t60_0_entry, "activate",
      G_CALLBACK(t60_0_callback), sg);
  g_signal_connect(sg->t60_pi_entry, "activate",
      G_CALLBACK(t60_pi_callback), sg);
  g_signal_connect(sg->bypass_t60_toggle, "toggled",
      G_CALLBACK(bypass_t60_callback), sg);
  g_signal_connect(sg->float_button, "toggled",
      G_CALLBACK(precision_callback), sg);
  g_signal_connect(sg->int_button, "toggled",
      G_CALLBACK(precision_callback), sg);
#ifdef ASM_IA32
  g_signal_connect(sg->asm_button, "toggled",
      G_CALLBACK(precision_callback), sg);
#endif
  g_signal_connect(sg->int_precision_scale, "format-value",
      G_CALLBACK(integer_precision_format_callback), 0);
  g_signal_connect(sg->int_precision_scale, "value-changed",
      G_CALLBACK(integer_precision_callback), sg);
  g_signal_connect(sg->impulse_drawing_area, "expose-event",
      G_CALLBACK(impulse_expose_callback), sg);
  g_signal_connect(sg->update_impulse_button, "clicked",
      G_CALLBACK(update_impulse_callback), sg);
  g_signal_connect(sg->impulse_length_scale, "format-value",
      G_CALLBACK(impulse_length_format_callback), 0);
  g_signal_connect(sg->impulse_play_button, "clicked",
      G_CALLBACK(impulse_play_callback), sg);
  g_signal_connect(sg->input_scale_entry, "activate",
      G_CALLBACK(input_scale_callback), sg);
  g_signal_connect(sg->output_scale_entry, "activate",
      G_CALLBACK(output_scale_callback), sg);

  preset_callback(0, sg);

  update_impulse_callback(GTK_BUTTON(sg->update_impulse_button), sg);

  /* ok */
  gtk_widget_show_all(sg->main_win);

  return;

list_error:
  fprintf(stderr, "you should not read this, contact" SVERB_EMAIL "\n");
  exit(1);
}

int main(int n, char **v)
{
  sverb s;
  sverb_gui sg;

  sverb_init(&s);

  gtk_init(&n, &v);

  sverb_gui_init(&sg, &s);

  gtk_main();

  return 0;
}
